from typing_extensions import deprecated

from zmq.constants import DeviceType
from zmq.error import *

from .context import *
from .frame import *
from .poll import *
from .socket import *
from .tracker import *
from .version import *

__all__ = [
    'device',
    'Context',
    'SyncContext',
    'DraftFDWarning',
    'ZMQBaseError',
    'ZMQBindError',
    'ZMQError',
    'NotDone',
    'ContextTerminated',
    'InterruptedSystemCall',
    'Again',
    'ZMQVersionError',
    'Frame',
    'Message',
    'Poller',
    'select',
    'Socket',
    'SyncSocket',
    'MessageTracker',
    '_FINISHED_TRACKER',
    'zmq_version',
    'zmq_version_info',
    'pyzmq_version',
    'pyzmq_version_info',
    '__version__',
    '__revision__',
    'Stopwatch',
]

@deprecated("use zmq.proxy instead")
def device(device_type: DeviceType, frontend: Socket, backend: Socket) -> int: ...

@deprecated("use time.monotonic() or time.perf_counter() instead")
class Stopwatch:
    def __init__(self) -> None: ...
    def start(self) -> None: ...
    def stop(self) -> int: ...



COPY_THRESHOLD: int
DRAFT_API: bool
__version__: str

from .backend import (
    IPC_PATH_MAX_LEN,
    PYZMQ_DRAFT_API,
    curve_keypair,
    curve_public,
    has,
    proxy,
    proxy_steerable,
    strerror,
    zmq_errno,
    zmq_poll,
)
from .constants import *
from .error import *
from .sugar import *

def get_includes() -> list[str]: ...
def get_library_dirs() -> list[str]: ...

__all__ = [
    'get_includes',
    'COPY_THRESHOLD',
    'DRAFT_API',
    'AFFINITY',
    'Again',
    'BACKLOG',
    'BINDTODEVICE',
    'BLOCKY',
    'BUSY_POLL',
    'CHANNEL',
    'CLIENT',
    'CONFLATE',
    'CONNECT_RID',
    'CONNECT_ROUTING_ID',
    'CONNECT_TIMEOUT',
    'CURVE',
    'CURVE_PUBLICKEY',
    'CURVE_SECRETKEY',
    'CURVE_SERVER',
    'CURVE_SERVERKEY',
    'Context',
    'ContextOption',
    'ContextTerminated',
    'DEALER',
    'DELAY_ATTACH_ON_CONNECT',
    'DGRAM',
    'DISCONNECT_MSG',
    'DISH',
    'DONTWAIT',
    'DeviceType',
    'DraftFDWarning',
    'EADDRINUSE',
    'EADDRNOTAVAIL',
    'EAFNOSUPPORT',
    'EAGAIN',
    'ECONNABORTED',
    'ECONNREFUSED',
    'ECONNRESET',
    'EFAULT',
    'EFSM',
    'EHOSTUNREACH',
    'EINPROGRESS',
    'EINVAL',
    'EMSGSIZE',
    'EMTHREAD',
    'ENETDOWN',
    'ENETRESET',
    'ENETUNREACH',
    'ENOBUFS',
    'ENOCOMPATPROTO',
    'ENOTCONN',
    'ENOTSOCK',
    'ENOTSUP',
    'EPROTONOSUPPORT',
    'ETERM',
    'ETIMEDOUT',
    'EVENTS',
    'EVENT_ACCEPTED',
    'EVENT_ACCEPT_FAILED',
    'EVENT_ALL',
    'EVENT_ALL_V1',
    'EVENT_ALL_V2',
    'EVENT_BIND_FAILED',
    'EVENT_CLOSED',
    'EVENT_CLOSE_FAILED',
    'EVENT_CONNECTED',
    'EVENT_CONNECT_DELAYED',
    'EVENT_CONNECT_RETRIED',
    'EVENT_DISCONNECTED',
    'EVENT_HANDSHAKE_FAILED_AUTH',
    'EVENT_HANDSHAKE_FAILED_NO_DETAIL',
    'EVENT_HANDSHAKE_FAILED_PROTOCOL',
    'EVENT_HANDSHAKE_SUCCEEDED',
    'EVENT_LISTENING',
    'EVENT_MONITOR_STOPPED',
    'EVENT_PIPES_STATS',
    'Enum',
    'Errno',
    'Event',
    'FAIL_UNROUTABLE',
    'FD',
    'FORWARDER',
    'Flag',
    'Frame',
    'GATHER',
    'GSSAPI',
    'GSSAPI_PLAINTEXT',
    'GSSAPI_PRINCIPAL',
    'GSSAPI_PRINCIPAL_NAMETYPE',
    'GSSAPI_SERVER',
    'GSSAPI_SERVICE_PRINCIPAL',
    'GSSAPI_SERVICE_PRINCIPAL_NAMETYPE',
    'HANDSHAKE_IVL',
    'HEARTBEAT_IVL',
    'HEARTBEAT_TIMEOUT',
    'HEARTBEAT_TTL',
    'HELLO_MSG',
    'HICCUP_MSG',
    'HWM',
    'IDENTITY',
    'IMMEDIATE',
    'INVERT_MATCHING',
    'IN_BATCH_SIZE',
    'IO_THREADS',
    'IPC_FILTER_GID',
    'IPC_FILTER_PID',
    'IPC_FILTER_UID',
    'IPC_PATH_MAX_LEN',
    'IPV4ONLY',
    'IPV6',
    'IntEnum',
    'IntFlag',
    'InterruptedSystemCall',
    'LAST_ENDPOINT',
    'LINGER',
    'LOOPBACK_FASTPATH',
    'MAXMSGSIZE',
    'MAX_MSGSZ',
    'MAX_SOCKETS',
    'MECHANISM',
    'METADATA',
    'MORE',
    'MSG_T_SIZE',
    'MULTICAST_HOPS',
    'MULTICAST_LOOP',
    'MULTICAST_MAXTPDU',
    'Message',
    'MessageOption',
    'MessageTracker',
    'NOBLOCK',
    'NORM_BLOCK_SIZE',
    'NORM_BUFFER_SIZE',
    'NORM_CC',
    'NORM_CCE',
    'NORM_CCE_ECNONLY',
    'NORM_CCL',
    'NORM_FIXED',
    'NORM_MODE',
    'NORM_NUM_AUTOPARITY',
    'NORM_NUM_PARITY',
    'NORM_PUSH',
    'NORM_SEGMENT_SIZE',
    'NORM_UNICAST_NACK',
    'NOTIFY_CONNECT',
    'NOTIFY_DISCONNECT',
    'NULL',
    'NormMode',
    'NotDone',
    'ONLY_FIRST_SUBSCRIBE',
    'OUT_BATCH_SIZE',
    'PAIR',
    'PEER',
    'PLAIN',
    'PLAIN_PASSWORD',
    'PLAIN_SERVER',
    'PLAIN_USERNAME',
    'POLLERR',
    'POLLIN',
    'POLLOUT',
    'POLLPRI',
    'PRIORITY',
    'PROBE_ROUTER',
    'PROTOCOL_ERROR_WS_UNSPECIFIED',
    'PROTOCOL_ERROR_ZAP_BAD_REQUEST_ID',
    'PROTOCOL_ERROR_ZAP_BAD_VERSION',
    'PROTOCOL_ERROR_ZAP_INVALID_METADATA',
    'PROTOCOL_ERROR_ZAP_INVALID_STATUS_CODE',
    'PROTOCOL_ERROR_ZAP_MALFORMED_REPLY',
    'PROTOCOL_ERROR_ZAP_UNSPECIFIED',
    'PROTOCOL_ERROR_ZMTP_CRYPTOGRAPHIC',
    'PROTOCOL_ERROR_ZMTP_INVALID_METADATA',
    'PROTOCOL_ERROR_ZMTP_INVALID_SEQUENCE',
    'PROTOCOL_ERROR_ZMTP_KEY_EXCHANGE',
    'PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_ERROR',
    'PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_HELLO',
    'PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_INITIATE',
    'PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_MESSAGE',
    'PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_READY',
    'PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_UNSPECIFIED',
    'PROTOCOL_ERROR_ZMTP_MALFORMED_COMMAND_WELCOME',
    'PROTOCOL_ERROR_ZMTP_MECHANISM_MISMATCH',
    'PROTOCOL_ERROR_ZMTP_UNEXPECTED_COMMAND',
    'PROTOCOL_ERROR_ZMTP_UNSPECIFIED',
    'PUB',
    'PULL',
    'PUSH',
    'PYZMQ_DRAFT_API',
    'PollEvent',
    'Poller',
    'QUEUE',
    'RADIO',
    'RATE',
    'RCVBUF',
    'RCVHWM',
    'RCVMORE',
    'RCVTIMEO',
    'RECONNECT_IVL',
    'RECONNECT_IVL_MAX',
    'RECONNECT_STOP',
    'RECONNECT_STOP_AFTER_DISCONNECT',
    'RECONNECT_STOP_CONN_REFUSED',
    'RECONNECT_STOP_HANDSHAKE_FAILED',
    'RECOVERY_IVL',
    'REP',
    'REQ',
    'REQ_CORRELATE',
    'REQ_RELAXED',
    'ROUTER',
    'ROUTER_BEHAVIOR',
    'ROUTER_HANDOVER',
    'ROUTER_MANDATORY',
    'ROUTER_NOTIFY',
    'ROUTER_RAW',
    'ROUTING_ID',
    'ReconnectStop',
    'RouterNotify',
    'SCATTER',
    'SERVER',
    'SHARED',
    'SNDBUF',
    'SNDHWM',
    'SNDMORE',
    'SNDTIMEO',
    'SOCKET_LIMIT',
    'SOCKS_PASSWORD',
    'SOCKS_PROXY',
    'SOCKS_USERNAME',
    'SRCFD',
    'STREAM',
    'STREAMER',
    'STREAM_NOTIFY',
    'SUB',
    'SUBSCRIBE',
    'SecurityMechanism',
    'Socket',
    'SocketOption',
    'SocketType',
    'Stopwatch',
    'SyncContext',
    'SyncSocket',
    'TCP_ACCEPT_FILTER',
    'TCP_KEEPALIVE',
    'TCP_KEEPALIVE_CNT',
    'TCP_KEEPALIVE_IDLE',
    'TCP_KEEPALIVE_INTVL',
    'TCP_MAXRT',
    'THREAD_AFFINITY_CPU_ADD',
    'THREAD_AFFINITY_CPU_REMOVE',
    'THREAD_NAME_PREFIX',
    'THREAD_PRIORITY',
    'THREAD_SAFE',
    'THREAD_SCHED_POLICY',
    'TOPICS_COUNT',
    'TOS',
    'TYPE',
    'UNSUBSCRIBE',
    'USE_FD',
    'VMCI_BUFFER_MAX_SIZE',
    'VMCI_BUFFER_MIN_SIZE',
    'VMCI_BUFFER_SIZE',
    'VMCI_CONNECT_TIMEOUT',
    'WSS_CERT_PEM',
    'WSS_HOSTNAME',
    'WSS_KEY_PEM',
    'WSS_TRUST_PEM',
    'WSS_TRUST_SYSTEM',
    'XPUB',
    'XPUB_MANUAL',
    'XPUB_MANUAL_LAST_VALUE',
    'XPUB_NODROP',
    'XPUB_VERBOSE',
    'XPUB_VERBOSER',
    'XPUB_WELCOME_MSG',
    'XREP',
    'XREQ',
    'XSUB',
    'XSUB_VERBOSE_UNSUBSCRIBE',
    'ZAP_DOMAIN',
    'ZAP_ENFORCE_DOMAIN',
    'ZMQBaseError',
    'ZMQBindError',
    'ZMQError',
    'ZMQVersionError',
    '_FINISHED_TRACKER',
    '__revision__',
    '__version__',
    'curve_keypair',
    'curve_public',
    'device',
    'has',
    'proxy',
    'proxy_steerable',
    'pyzmq_version',
    'pyzmq_version_info',
    'select',
    'strerror',
    'zmq_errno',
    'zmq_poll',
    'zmq_version',
    'zmq_version_info',
]



from collections.abc import Sequence
from typing import Any, Final, Protocol, overload, type_check_only

from _typeshed import HasFileno
from typing_extensions import Buffer, Literal, Self

import zmq

from .select import select_backend as select_backend

__all__ = [
    'Context',
    'Socket',
    'Frame',
    'Message',
    'proxy',
    'proxy_steerable',
    'zmq_poll',
    'strerror',
    'zmq_errno',
    'has',
    'curve_keypair',
    'curve_public',
    'zmq_version_info',
    'IPC_PATH_MAX_LEN',
    'PYZMQ_DRAFT_API',
]

IPC_PATH_MAX_LEN: Final[int] = ...
PYZMQ_DRAFT_API: Final[bool] = ...

class Frame:
    more: bool
    tracker: Any

    def __init__(
        self,
        data: Any = None,
        track: bool = False,
        copy: bool | None = None,
        copy_threshold: int | None = None,
    ) -> None: ...
    def __len__(self) -> int: ...
    def __copy__(self) -> Self: ...
    def fast_copy(self) -> Self: ...

    #
    @overload
    def get(self, option: int | Literal["routing_id"]) -> int: ...
    @overload
    def get(self, option: bytes | Literal["group"]) -> str: ...
    @overload
    def get(self, option: str) -> int | str: ...

    #
    @overload
    def set(self, option: Literal["routing_id"], value: int) -> None: ...
    @overload
    def set(self, option: Literal["group"], value: str | bytes) -> None: ...
    @overload
    def set(self, option: int, value: int) -> None: ...

    #
    @property
    def buffer(self) -> memoryview: ...
    @property
    def bytes(self) -> bytes: ...

Message = Frame

class Socket:
    context: zmq.Context
    copy_threshold: int

    # specific option types
    FD: int

    def __init__(
        self,
        context: Context | None = None,
        socket_type: int = -1,
        shadow: int = 0,
        copy_threshold: int | None = None,
    ) -> None: ...

    #
    @property
    def underlying(self) -> int: ...
    @property
    def closed(self) -> bool: ...

    #
    def close(self, linger: int | None = None) -> None: ...

    #
    def set(self, option: int, optval: int | bytes) -> None: ...
    def get(self, option: int) -> int | bytes: ...

    #
    def bind(self, addr: str) -> None: ...
    def connect(self, addr: str) -> None: ...
    def unbind(self, addr: str | bytes) -> None: ...
    def disconnect(self, addr: str | bytes) -> None: ...
    def monitor(
        self,
        addr: str | bytes | None,
        events: int = zmq.EVENT_ALL,
    ) -> None: ...
    def join(self, group: str | bytes) -> None: ...
    def leave(self, group: str | bytes) -> None: ...

    #
    @overload  # data: Buffer, copy=True (default)
    def send(
        self,
        data: Buffer,
        flags: int = 0,
        copy: Literal[True] = True,
        track: bool = False,
    ) -> None: ...
    @overload  # data: Buffer, copy=False (keyword)
    def send(
        self,
        data: Buffer,
        flags: int = 0,
        *,
        copy: Literal[False],
        track: bool = False,
    ) -> zmq.MessageTracker: ...
    @overload  # data: Buffer, copy=False (positional)
    def send(
        self,
        data: Buffer,
        flags: int,
        copy: Literal[False],
        track: bool = False,
    ) -> zmq.MessageTracker: ...
    @overload  # data: Frame
    def send(
        self,
        data: Frame,
        flags: int = 0,
        copy: bool = True,
        track: bool = False,
    ) -> zmq.MessageTracker: ...

    #
    @overload  # copy=True (default)
    def recv(
        self,
        flags: int = 0,
        copy: Literal[True] = True,
        track: bool = False,
    ) -> bytes: ...
    @overload  # copy=False (keyword)
    def recv(
        self,
        flags: int = 0,
        *,
        copy: Literal[False],
        track: bool = False,
    ) -> zmq.Frame: ...
    @overload  # copy=False (positional)
    def recv(
        self,
        flags: int,
        copy: Literal[False],
        track: bool = False,
    ) -> zmq.Frame: ...
    @overload  # fallback overload (mypy bug workaround)
    def recv(
        self,
        flags: int = 0,
        copy: bool = True,
        track: bool = False,
    ) -> bytes | zmq.Frame: ...

    #
    def recv_into(
        self, buffer: Buffer, /, *, nbytes: int = 0, flags: int = 0
    ) -> int: ...

class Context:
    handle: int
    closed: bool

    def __init__(self, io_threads: int = 1, shadow: int = 0) -> None: ...

    #
    @property
    def underlying(self) -> int: ...

    #
    def term(self) -> None: ...
    def set(self, option: int, optval: int) -> None: ...
    def get(self, option: int) -> int: ...

def zmq_errno() -> int: ...
def strerror(errno: int) -> str: ...
def zmq_version_info() -> tuple[int, int, int]: ...
def has(capability: str) -> bool: ...
def curve_keypair() -> tuple[bytes, bytes]: ...
def curve_public(secret_key: str | bytes) -> bytes: ...

#
@overload
def zmq_poll(
    sockets: Sequence[tuple[Socket, int]],
    timeout: int = -1,
) -> list[tuple[Socket, int]]: ...
@overload
def zmq_poll(
    sockets: Sequence[tuple[int | HasFileno, int]],
    timeout: int = -1,
) -> list[tuple[int, int]]: ...

#
def proxy(frontend: Socket, backend: Socket, capture: Socket | None = None) -> int: ...
def proxy_steerable(
    frontend: Socket,
    backend: Socket,
    capture: Socket | None = None,
    control: Socket | None = None,
) -> int: ...

@type_check_only
class _MonitoredQueueFunction(Protocol):
    def __call__(
        self,
        /,
        in_socket: Socket,
        out_socket: Socket,
        mon_socket: Socket,
        in_prefix: Buffer = b"in",
        out_prefix: Buffer = b"out",
    ) -> int: ...

# None for the cffi backend
monitored_queue: Final[_MonitoredQueueFunction | None] = ...



"""Python implementation of Z85 85-bit encoding

Z85 encoding is a plaintext encoding for a bytestring interpreted as 32bit integers.
Since the chunks are 32bit, a bytestring must be a multiple of 4 bytes.
See ZMQ RFC 32 for details.


"""

# Copyright (C) PyZMQ Developers
# Distributed under the terms of the Modified BSD License.
from __future__ import annotations

import struct
from typing import Final, Union

_RawBytes = Union[bytes, bytearray, memoryview]

# Z85CHARS is the base 85 symbol table
Z85CHARS: Final = b"0123456789abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.-:+=^!/*?&<>()[]{}@%$#"
# Z85MAP maps integers in [0,84] to the appropriate character in Z85CHARS
Z85MAP: Final[dict[int, int]] = {c: idx for idx, c in enumerate(Z85CHARS)}

_85s = [85**i for i in range(5)][::-1]


def encode(rawbytes: _RawBytes) -> bytes:
    """encode raw bytes into Z85"""
    # Accepts only byte arrays bounded to 4 bytes
    if len(rawbytes) % 4:
        raise ValueError(f"length must be multiple of 4, not {len(rawbytes)}")

    nvalues = len(rawbytes) // 4
    values = struct.unpack(f'>{nvalues:d}I', rawbytes)
    encoded = []
    for v in values:
        for offset in _85s:
            encoded.append(Z85CHARS[(v // offset) % 85])

    return bytes(encoded)


def decode(z85bytes: str | _RawBytes) -> bytes:
    """decode Z85 bytes to raw bytes, accepts ASCII string"""
    if isinstance(z85bytes, str):
        try:
            z85bytes = z85bytes.encode('ascii')
        except UnicodeEncodeError:
            raise ValueError('string argument should contain only ASCII characters')

    if len(z85bytes) % 5:
        raise ValueError(f"Z85 length must be multiple of 5, not {len(z85bytes)}")

    nvalues = len(z85bytes) // 5
    values = []
    for i in range(0, len(z85bytes), 5):
        value = 0
        for j, offset in enumerate(_85s):
            value += Z85MAP[z85bytes[i + j]] * offset
        values.append(value)
    return struct.pack(f'>{nvalues:d}I', *values)



"""
Generate a zmq curve keypair.

Generates a 'public' and a 'secret' key.
Add `--json` to output as a JSON dictionary.
Add `--secret SECRET` to derive public key from an existing secret key.

CURVE keys are always z85-encoded (ASCII safe).
"""

import argparse
import json
import sys

import zmq


def _main() -> None:
    parser = argparse.ArgumentParser(
        prog="python3 -m zmq.curve_keygen",
        description=__doc__,
    )
    parser.add_argument("--json", action="store_true", help="output keys as JSON dict")
    parser.add_argument(
        "--secret", type=str, help="specify secret key to derive a public key"
    )
    opts = parser.parse_args()
    if opts.secret:
        secret = opts.secret
        public = zmq.curve_public(secret.encode()).decode()
    else:
        public_b, secret_b = zmq.curve_keypair()
        public = public_b.decode()
        secret = secret_b.decode()
    if opts.json:
        json.dump(
            {
                "public": public,
                "secret": secret,
            },
            sys.stdout,
            indent=1,
            sort_keys=True,
        )
        sys.stdout.write("\n")
    else:
        # stdout, match libzmq curve_keygen format
        # without the big disclaimer
        print("== CURVE PUBLIC KEY ==")
        print(public)
        print()
        print("== CURVE SECRET KEY ==")
        print(secret)


if __name__ == "__main__":
    _main()
